#!/usr/bin/env python3
import os
import platform
import sys
import tempfile

import click

from .layout import KeyboardLayout
from .xkb_manager import XKBManager


@click.group()
def cli():
    if platform.system() != 'Linux':
        sys.exit('This command is only compatible with GNU/Linux, sorry.')


@cli.command()
@click.argument('input', nargs=1, type=click.Path(exists=True))
def apply(input):
    """ Apply a Kalamine layout. """

    layout = KeyboardLayout(input)
    with tempfile.NamedTemporaryFile(mode='w+', suffix='.xkb', encoding='utf-8') as temp_file:
        try:
            temp_file.write(layout.xkb)
            os.system(f"xkbcomp -w0 {temp_file.name} $DISPLAY")
        finally:
            temp_file.close()


@cli.command()
@click.argument('layouts', nargs=-1, type=click.Path(exists=True))
def install(layouts):
    """ Install a list of Kalamine layouts. """
    if len(layouts) == 0:
        return

    xkb = XKBManager()
    for file in layouts:
        xkb.add(KeyboardLayout(file))
    index = xkb.index
    xkb.update()

    print()
    print(f"Successfully installed. You can try the layout{'s' if len(layouts) > 1 else ''} with:")
    for locale, variants in index:
        for name in variants.keys():
            print(f"    setxkbmap {locale} -variant {name}")
    print()


@cli.command()
def clean():
    """ Clean all installed Kalamine layouts: drop the obsolete 'type' attr. """

    xkb = XKBManager()
    xkb.clean()


@cli.command()
@click.argument('mask', default='*')
@click.option('--all', '-a', is_flag=True)
def list(mask, all):
    """ List all installed Kalamine layouts. """

    xkb = XKBManager()
    layouts = xkb.list_all(mask) if all else xkb.list(mask)
    for locale, variants in sorted(layouts.items()):
        for name, desc in sorted(variants.items()):
            id = f"{locale}/{name}"
            print(f"{id:<24}   {desc}")


@cli.command()
@click.argument('mask')  # [locale]/[name]
def remove(mask):
    """ Remove an existing Kalamine layout. """

    xkb = XKBManager()
    for locale, variants in xkb.list(mask).items():
        for name in variants.keys():
            xkb.remove(locale, name)
    xkb.update()



#!/usr/bin/env python3
import os
import yaml


def lines_to_text(lines, indent=''):
    out = ''
    for line in lines:
        if len(line):
            out += indent + line
        out += '\n'
    return out[:-1]


def text_to_lines(text):
    return text.split('\n')


def open_local_file(file_name):
    return open(os.path.join(os.path.dirname(__file__), file_name), encoding='utf-8')


def load_data(filename):
    return yaml.load(open_local_file(os.path.join('data', filename)),
                     Loader=yaml.SafeLoader)


DEAD_KEYS = load_data('dead_keys.yaml')
ODK_ID = '**'  # must match the value in dead_keys.yaml
LAYER_KEYS = [
    '- Digits',
    'ae01', 'ae02', 'ae03', 'ae04', 'ae05',
    'ae06', 'ae07', 'ae08', 'ae09', 'ae10',

    '- Letters, first row',
    'ad01', 'ad02', 'ad03', 'ad04', 'ad05',
    'ad06', 'ad07', 'ad08', 'ad09', 'ad10',

    '- Letters, second row',
    'ac01', 'ac02', 'ac03', 'ac04', 'ac05',
    'ac06', 'ac07', 'ac08', 'ac09', 'ac10',

    '- Letters, third row',
    'ab01', 'ab02', 'ab03', 'ab04', 'ab05',
    'ab06', 'ab07', 'ab08', 'ab09', 'ab10',

    '- Pinky keys',
    'ae11', 'ae12', 'ae13', 'ad11', 'ad12',
    'ac11', 'ab11', 'tlde', 'bksl', 'lsgt',

    '- Space bar',
    'spce'
]



#!/usr/bin/env python3
import json
import os
import pkg_resources

import click

from .layout import KeyboardLayout


def pretty_json(layout, path):
    """ Pretty-prints the JSON layout. """
    text = json.dumps(layout.json, indent=2, ensure_ascii=False) \
               .replace('\n      ', ' ') \
               .replace('\n    ]', ' ]') \
               .replace('\n    }', ' }')
    with open(path, 'w', encoding='utf8') as file:
        file.write(text)


def make_all(layout, subdir):
    def out_path(ext=''):
        return os.path.join(subdir, layout.meta['fileName'] + ext)

    if not os.path.exists(subdir):
        os.makedirs(subdir)

    # Windows driver
    klc_path = out_path('.klc')
    with open(klc_path, 'w', encoding='utf-16le', newline='\r\n') as file:
        file.write(layout.klc)
    print('... ' + klc_path)

    # macOS driver
    osx_path = out_path('.keylayout')
    with open(osx_path, 'w', encoding='utf-8', newline='\n') as file:
        file.write(layout.keylayout)
    print('... ' + osx_path)

    # Linux driver, user-space
    xkb_path = out_path('.xkb')
    with open(xkb_path, 'w', encoding='utf-8', newline='\n') as file:
        file.write(layout.xkb)
    print('... ' + xkb_path)

    # Linux driver, user-space
    xkb_custom_path = out_path('.xkb_custom')
    with open(xkb_custom_path, 'w', encoding='utf-8', newline='\n') as file:
        file.write(layout.xkb_patch)
    print('... ' + xkb_custom_path)

    # JSON data
    json_path = out_path('.json')
    pretty_json(layout, json_path)
    print('... ' + json_path)


@click.command()
@click.argument('input', nargs=-1, type=click.Path(exists=True))
@click.option('--version', '-v', is_flag=True)
@click.option('--out',
              default='all',
              type=click.Path(),
              help='Keyboard driver(s) to generate.')
def make(input, version, out):
    """ Convert toml/yaml descriptions into OS-specific keyboard layouts. """

    if version:
        print(f"kalamine { pkg_resources.require('kalamine')[0].version }")

    for input_file in input:
        layout = KeyboardLayout(input_file)

        # default: build all in the `dist` subdirectory
        if out == 'all':
            make_all(layout, 'dist')
            continue

        # quick output: reuse the input name and change the file extension
        if out in ['keylayout', 'klc', 'xkb', 'xkb_custom']:
            output_file = os.path.splitext(input_file)[0] + '.' + out
        else:
            output_file = out

        # detailed output
        if output_file.endswith('.klc'):
            with open(output_file, 'w', encoding='utf-16le', newline='\r\n') as file:
                file.write(layout.klc)
        elif output_file.endswith('.keylayout'):
            with open(output_file, 'w', encoding='utf-8', newline='\n') as file:
                file.write(layout.keylayout)
        elif output_file.endswith('.xkb'):
            with open(output_file, 'w', encoding='utf-8', newline='\n') as file:
                file.write(layout.xkb)
        elif output_file.endswith('.xkb_custom'):
            with open(output_file, 'w', encoding='utf-8', newline='\n') as file:
                file.write(layout.xkb_patch)
        elif output_file.endswith('.json'):
            pretty_json(layout, output_file)
        else:
            print('Unsupported output format.')
            return

        # successfully converted, display file name
        print('... ' + output_file)


if __name__ == '__main__':
    make()



#!/usr/bin/env python3
from .utils import load_data, LAYER_KEYS, ODK_ID


###
# Helpers
#

KEY_CODES = load_data('key_codes.yaml')
XKB_KEY_SYM = load_data('key_sym.yaml')


def hex_ord(char):
    return hex(ord(char))[2:].zfill(4)


def xml_proof(char):
    if char not in '<&"\u0020\u00a0\u202f>':
        return char
    return f"&#x{hex_ord(char)};"


def xml_proof_id(symbol):
    return symbol[2:-1] if symbol.startswith('&#x') else symbol


###
# GNU/Linux: XKB
# - standalone xkb file to be used by `setxkbcomp` (Xorg only)
# - system-wide installer script for Xorg & Wayland
#

def xkb_keymap(layout, eight_levels):
    """ Linux layout. """

    show_description = True
    max_length = 16  # `ISO_Level3_Latch` should be the longest symbol name

    output = []
    for key_name in LAYER_KEYS:
        if key_name.startswith('-'):  # separator
            if output:
                output.append('')
            output.append('//' + key_name[1:])
            continue

        symbols = []
        description = ' //'
        for layer in layout.layers:
            if key_name in layer:
                symbol = layer[key_name]
                desc = symbol
                if symbol in layout.dead_keys:
                    dk = layout.dead_keys[symbol]
                    desc = dk['alt_self']
                    if dk['char'] == ODK_ID:
                        symbol = 'ISO_Level3_Latch'
                    else:
                        symbol = 'dead_' + dk['name']
                elif symbol in XKB_KEY_SYM \
                        and len(XKB_KEY_SYM[symbol]) <= max_length:
                    symbol = XKB_KEY_SYM[symbol]
                else:
                    symbol = 'U' + hex_ord(symbol).upper()
            else:
                desc = ' '
                symbol = 'VoidSymbol'

            description += ' ' + desc
            symbols.append(symbol.ljust(max_length))

        key = 'key <{}> {{[ {}, {}, {}, {}]}};'  # 4-level layout by default
        if layout.has_altgr and layout.has_1dk:
            # 6 layers are needed: they won't fit on the 4-level format.
            # System XKB files require a Neo-like eight-level solution.
            # Standalone XKB files work best with a dual-group solution:
            # one 4-level group for base+1dk, one two-level group for AltGr.
            if eight_levels:  # system XKB file (patch)
                key = 'key <{}> {{[ {}, {}, {}, {}, {}, {}, {}, {}]}};'
                symbols.append('VoidSymbol'.ljust(max_length))
                symbols.append('VoidSymbol'.ljust(max_length))
            else:  # user-space XKB file (standalone)
                key = 'key <{}> {{[ {}, {}, {}, {}],[ {}, {}]}};'
        elif layout.has_altgr:
            del symbols[3]
            del symbols[2]

        line = key.format(* [key_name.upper()] + symbols)
        if show_description:
            line += description.rstrip()
            if line.endswith('\\'):
                line += ' '  # escape trailing backslash
        output.append(line)

    return output


###
# Windows: KLC
# To be used by the MS Keyboard Layout Creator to generate an installer.
# https://www.microsoft.com/en-us/download/details.aspx?id=102134
# https://levicki.net/articles/2006/09/29/HOWTO_Build_keyboard_layouts_for_Windows_x64.php
# Also supported by KbdEdit: http://www.kbdedit.com/ (non-free).
# Note: blank lines and comments in KLC sections are removed from the output
# file because they are not recognized by KBDEdit (as of v19.8.0).
#

def klc_keymap(layout):
    """ Windows layout, main part. """

    supported_symbols = \
        '1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'

    output = []
    for key_name in LAYER_KEYS:
        if key_name.startswith('-'):
            continue

        symbols = []
        description = '//'
        alpha = False

        for i in [0, 1, 4, 5]:
            layer = layout.layers[i]

            if key_name in layer:
                symbol = layer[key_name]
                desc = symbol
                if symbol in layout.dead_keys:
                    desc = layout.dead_keys[symbol]['alt_space']
                    symbol = hex_ord(desc) + '@'
                else:
                    if i == 0:
                        alpha = symbol.upper() != symbol
                    if symbol not in supported_symbols:
                        symbol = hex_ord(symbol)
                symbols.append(symbol)
            else:
                desc = ' '
                symbols.append('-1')
            description += ' ' + desc

        if layout.has_altgr:
            output.append('\t'.join([
                KEY_CODES['klc'][key_name],    # scan code & virtual key
                '1' if alpha else '0',         # affected by CapsLock?
                symbols[0], symbols[1], '-1',  # base layer
                symbols[2], symbols[3],        # altgr layer
                description.strip()
            ]))
        else:
            output.append('\t'.join([
                KEY_CODES['klc'][key_name],    # scan code & virtual key
                '1' if alpha else '0',         # affected by CapsLock?
                symbols[0], symbols[1], '-1',  # base layer
                description.strip()
            ]))

    return output


def klc_deadkeys(layout):
    """ Windows layout, dead keys. """

    output = []

    def append_line(base, alt):
        output.append(f"{hex_ord(base)}\t{hex_ord(alt)}\t// {base} -> {alt}")

    for k in layout.dk_index:
        dk = layout.dead_keys[k]

        output.append('// DEADKEY: ' + dk['name'].upper() + ' //{{{')
        output.append('DEADKEY\t' + hex_ord(dk['alt_space']))

        if k == ODK_ID:
            output.extend(klc_1dk(layout))
        else:
            for i in range(len(dk['base'])):
                append_line(dk['base'][i], dk['alt'][i])

        append_line('\u00a0', dk['alt_space'])
        append_line('\u0020', dk['alt_space'])

        output.append('//}}}')
        output.append('')

    return output[:-1]


def klc_dk_index(layout):
    """ Windows layout, dead key index. """

    output = []
    for k in layout.dk_index:
        dk = layout.dead_keys[k]
        output.append(f"{hex_ord(dk['alt_space'])}\t\"{dk['name'].upper()}\"")
    return output


def klc_1dk(layout):
    """ Windows layout, 1dk. """

    output = []
    for i in [0, 1]:
        base_layer = layout.layers[i]
        ext_layer = layout.layers[i + 2]

        for key_name in LAYER_KEYS:
            if key_name.startswith('- Space') or key_name == 'spce':
                continue

            if key_name.startswith('-'):
                if output:
                    output.append('')
                output.append('//' + key_name[1:])
                continue

            if key_name in ext_layer:
                base = base_layer[key_name]
                if base in layout.dead_keys:
                    base = layout.dead_keys[base]['alt_space']
                ext = ext_layer[key_name]
                if ext in layout.dead_keys:
                    ext = layout.dead_keys[ext]['alt_space']
                    odk = hex_ord(ext) + '@'
                else:
                    odk = hex_ord(ext)

                output.append('\t'.join([
                    hex_ord(base), odk, '// ' + base + ' -> ' + ext
                ]))

    return output


###
# macOS: keylayout
# https://developer.apple.com/library/content/technotes/tn2056/
#

def osx_keymap(layout):
    """ macOS layout, main part. """

    ret_str = []
    for index in range(5):
        layer = layout.layers[[0, 1, 0, 4, 5][index]]
        caps = index == 2

        def has_dead_keys(letter):
            if letter in '\u0020\u00a0\u202f':  # space
                return True
            for k in layout.dead_keys:
                if letter in layout.dead_keys[k]['base']:
                    return True
            return False

        output = []
        for key_name in LAYER_KEYS:
            if key_name in ['ae13', 'ab11']:  # ABNT / JIS keys
                continue  # these two keys are not supported yet

            if key_name.startswith('-'):
                if output:
                    output.append('')
                output.append('<!--' + key_name[1:] + ' -->')
                continue

            symbol = '&#x0010;'
            final_key = True

            if key_name in layer:
                key = layer[key_name]
                if key in layout.dead_keys:
                    symbol = 'dead_' + layout.dead_keys[key]['name']
                    final_key = False
                else:
                    symbol = xml_proof(key.upper() if caps else key)
                    final_key = not has_dead_keys(key.upper())

            char = f"code=\"{KEY_CODES['osx'][key_name]}\"".ljust(10)
            if final_key:
                action = f"output=\"{symbol}\""
            else:
                action = f"action=\"{xml_proof_id(symbol)}\""
            output.append(f"<key {char} {action} />")

        ret_str.append(output)
    return ret_str


def osx_actions(layout):
    """ macOS layout, dead key actions. """

    ret_actions = []

    def when(state, action):
        state_attr = f"state=\"{state}\"".ljust(18)
        if action in layout.dead_keys:
            action_attr = f"next=\"{layout.dead_keys[action]['name']}\""
        elif action.startswith('dead_'):
            action_attr = f"next=\"{action[5:]}\""
        else:
            action_attr = f"output=\"{xml_proof(action)}\""
        return f"  <when {state_attr} {action_attr} />"

    def append_actions(symbol, actions):
        ret_actions.append(f"<action id=\"{xml_proof_id(symbol)}\">")
        ret_actions.append(when('none', symbol))
        for (state, out) in actions:
            ret_actions.append(when(state, out))
        ret_actions.append('</action>')

    # dead key definitions
    for key in layout.dead_keys:
        symbol = layout.dead_keys[key]['name']
        ret_actions.append(f"<action id=\"dead_{symbol}\">")
        ret_actions.append(f"  <when state=\"none\" next=\"{symbol}\" />")
        ret_actions.append('</action>')
        continue

    # normal key actions
    for key_name in LAYER_KEYS:
        if key_name.startswith('-'):
            ret_actions.append('')
            ret_actions.append('<!--' + key_name[1:] + ' -->')
            continue

        for i in [0, 1]:
            if key_name not in layout.layers[i]:
                continue

            key = layout.layers[i][key_name]
            if i and key == layout.layers[0][key_name]:
                continue
            if key in layout.dead_keys:
                continue

            actions = []
            for k in layout.dk_index:
                dk = layout.dead_keys[k]
                if key in dk['base']:
                    idx = dk['base'].index(key)
                    actions.append((dk['name'], dk['alt'][idx]))
            if actions:
                append_actions(xml_proof(key), actions)

    # spacebar actions
    actions = []
    for k in layout.dk_index:
        dk = layout.dead_keys[k]
        actions.append((dk['name'], dk['alt_space']))
    append_actions('&#x0020;', actions)
    append_actions('&#x00a0;', actions)
    append_actions('&#x202f;', actions)

    return ret_actions


def osx_terminators(layout):
    """ macOS layout, dead key terminators. """

    ret_terminators = []
    for k in layout.dk_index:
        dk = layout.dead_keys[k]
        state = f"state=\"{dk['name']}\"".ljust(18)
        output = f"output=\"{xml_proof(dk['alt_self'])}\""
        ret_terminators.append(f"<when {state} {output} />")
    return ret_terminators


###
# Web: JSON
# To be used with the <x-keyboard> web component.
# https://github.com/fabi1cazenave/x-keyboard
#

def web_keymap(layout):
    """ Web layout, main part. """

    keymap = {}
    for key_name in LAYER_KEYS:
        if key_name.startswith('-'):
            continue
        chars = list('')
        for i in [0, 1, 4, 5]:
            if key_name in layout.layers[i]:
                chars.append(layout.layers[i][key_name])
        if chars:
            keymap[KEY_CODES['web'][key_name]] = chars

    return keymap


def web_deadkeys(layout):
    """ Web layout, dead keys. """

    deadkeys = {}
    if layout.has_1dk:  # ensure 1dk is first in the dead key dictionary
        deadkeys[ODK_ID] = {}
    for (id, dk) in layout.dead_keys.items():
        deadkeys[id] = {}
        deadkeys[id][id] = dk['alt_self']
        deadkeys[id]['\u0020'] = dk['alt_space']
        deadkeys[id]['\u00a0'] = dk['alt_space']
        deadkeys[id]['\u202f'] = dk['alt_space']
        if id == ODK_ID:
            for key_name in LAYER_KEYS:
                if key_name.startswith('-'):
                    continue
                for i in [3, 2]:
                    if key_name in layout.layers[i]:
                        deadkeys[id][layout.layers[i - 2][key_name]] = \
                            layout.layers[i][key_name]
        else:
            base = dk['base']
            alt = dk['alt']
            for i in range(len(base)):
                deadkeys[id][base[i]] = alt[i]

    return deadkeys



#!/usr/bin/env python3
import os
import sys
import traceback

from lxml import etree
from lxml.builder import E


class XKBManager:
    """ Wrapper to list/add/remove keyboard drivers to XKB. """

    def __init__(self, xkb_root='/usr/share/X11/xkb/'):
        self._rootdir = xkb_root
        self._index = {}

    @property
    def index(self):
        return self._index.items()

    def add(self, layout):
        locale = layout.meta['locale']
        variant = layout.meta['variant']
        if locale not in self._index:
            self._index[locale] = {}
        self._index[locale][variant] = layout

    def remove(self, locale, variant):
        if locale not in self._index:
            self._index[locale] = {}
        self._index[locale][variant] = None

    def update(self):
        update_symbols(self._rootdir, self._index)  # XKB/symbols/{locales}
        update_rules(self._rootdir, self._index)  # XKB/rules/{base,evdev}.xml
        self._index = {}

    def clean(self):
        clean_rules(self._rootdir)  # XKB/rules/{base,evdev}.xml

    def list(self, mask=''):
        layouts = list_rules(self._rootdir, mask)
        return list_symbols(self._rootdir, layouts)

    def list_all(self, mask=''):
        return list_rules(self._rootdir, mask)


""" On GNU/Linux, keyboard layouts must be installed in /usr/share/X11/xkb. To
    be able to revert a layout installation, Kalamine marks layouts like this:

    - XKB/symbols/[locale]: layout definitions
        // KALAMINE::[NAME]::BEGIN
        xkb_symbols "[name]" { ... }
        // KALAMINE::[NAME]::END

    Earlier versions of XKalamine used to mark index files as well but recent
    versions of Gnome do not support the custom `type` attribute any more, which
    must be removed:

    - XKB/rules/{base,evdev}.xml: layout references
        <variant type="kalamine">
            <configItem>
                <name>lafayette42</name>
                <description>French (Lafayette42)</description>
            </configItem>
        </variant>

    Even worse, the Lafayette project has released a first installer before
    the XKalamine installer was developed, so we have to handle this situation
    too:

    - XKB/symbols/[locale]: layout definitions
        // LAFAYETTE::BEGIN
        xkb_symbols "lafayette"   { ... }
        xkb_symbols "lafayette42" { ... }
        // LAFAYETTE::END

    - XKB/rules/{base,evdev}.xml: layout references
        <variant type="lafayette">
            <configItem>
                <name>lafayette</name>
                <description>French (Lafayette)</description>
            </configItem>
        </variant>
        <variant type="lafayette">
            <configItem>
                <name>lafayette42</name>
                <description>French (Lafayette42)</description>
            </configItem>
        </variant>

    Consequence: these two Lafayette layouts must be uninstalled together.
    Because of the way they are grouped in symbols/fr, it is impossible to
    remove one without removing the other.
"""

def clean_legacy_lafayette():
    return


###############################################################################
# Helpers: XKB/symbols
#


LEGACY_MARK = {
    'begin': '// LAFAYETTE::BEGIN\n',
    'end': '// LAFAYETTE::END\n'
}


def get_symbol_mark(name):
    return {
        'begin': '// KALAMINE::' + name.upper() + '::BEGIN\n',
        'end': '// KALAMINE::' + name.upper() + '::END\n'
    }


def is_new_symbol_mark(line):
    if line.endswith('::BEGIN\n'):
        if line.startswith('// KALAMINE::'):
            return line[13:-8].lower()  # XXX Kalamine expects lowercase names
        elif line.startswith('// LAFAYETTE::'):  # obsolete marker
            return 'lafayette'
    return None


def update_symbols_locale(path, named_layouts):
    """ Update Kalamine layouts in an xkb/symbols file. """

    text = ''
    modified_text = False
    with open(path, 'r+', encoding='utf-8') as symbols:

        # look for Kalamine layouts to be updated or removed
        between_marks = False
        closing_mark = ''
        for line in symbols:
            name = is_new_symbol_mark(line)
            if name:
                if name in named_layouts.keys():
                    closing_mark = line[:-6] + 'END\n'
                    modified_text = True
                    between_marks = True
                    text = text.rstrip()
                else:
                    text += line
            elif line.endswith('::END\n'):
                if between_marks and line.startswith(closing_mark):
                    between_marks = False
                    closing_mark = ''
                else:
                    text += line
            elif not between_marks:
                text += line

        # clear previous Kalamine layouts if needed
        if modified_text:
            symbols.seek(0)
            symbols.write(text.rstrip() + '\n')
            symbols.truncate()

        # add new Kalamine layouts
        for name, layout in named_layouts.items():
            if layout is None:
                print('      - ' + name)
            else:
                print('      + ' + name)
                mark = get_symbol_mark(name)
                symbols.write('\n')
                symbols.write(mark['begin'])
                symbols.write(layout.xkb_patch.rstrip() + '\n')
                symbols.write(mark['end'])

        symbols.close()


def update_symbols(xkb_root, kb_index):
    """ Update Kalamine layouts in all xkb/symbols files. """

    for locale, named_layouts in kb_index.items():
        path = os.path.join(xkb_root, 'symbols', locale)
        if not os.path.exists(path):
            exit_LocaleNotSupported(locale)

        try:
            print('... ' + path)
            update_symbols_locale(path, named_layouts)

        except Exception as exc:
            exit_FileNotWritable(exc, path)


def list_symbols(xkb_root, kb_index):
    """ Filter input layouts: only keep the ones defined with Kalamine. """

    filtered_index = {}
    for locale, variants in sorted(kb_index.items()):
        path = os.path.join(xkb_root, 'symbols', locale)
        if not os.path.exists(path):
            continue

        with open(path, 'r', encoding='utf-8') as symbols:
            for line in symbols:
                name = is_new_symbol_mark(line)
                if name in variants.keys():
                    if locale not in filtered_index:
                        filtered_index[locale] = {}
                    filtered_index[locale][name] = variants[name]

    return filtered_index


###############################################################################
# Helpers: XKB/rules
#

def get_rules_locale(tree, locale):
    query = f"//layout/configItem/name[text()=\"{locale}\"]/../.."
    result = tree.xpath(query)
    if len(result) != 1:
        exit_LocaleNotSupported(locale)
    return tree.xpath(query)[0]


def remove_rules_variant(variant_list, name):
    query = f"variant/configItem/name[text()=\"{name}\"]/../.."
    for variant in variant_list.xpath(query):
        variant.getparent().remove(variant)


def add_rules_variant(variant_list, name, description):
    variant_list.append(
        E.variant(
            E.configItem(
                E.name(name),
                E.description(description))))


def update_rules(xkb_root, kb_index):
    """ Update references in XKB/rules/{base,evdev}.xml. """

    for filename in ['base.xml', 'evdev.xml']:
        try:
            path = os.path.join(xkb_root, 'rules', filename)
            tree = etree.parse(path, etree.XMLParser(remove_blank_text=True))

            for locale, named_layouts in kb_index.items():
                vlist = get_rules_locale(tree, locale).xpath('variantList')
                if len(vlist) != 1:
                    exit(f"Error: unexpected xml format in {path}.")
                for name, layout in named_layouts.items():
                    remove_rules_variant(vlist[0], name)
                    if layout is not None:
                        description = layout.meta['description']
                        add_rules_variant(vlist[0], name, description)

            tree.write(path, pretty_print=True, xml_declaration=True,
                       encoding='utf-8')
            print('... ' + path)

        except Exception as exc:
            exit_FileNotWritable(exc, path)


def clean_rules(xkb_root):
    """ Drop the obsolete 'type' attributes Kalamine used to add. """

    for filename in ['base.xml', 'evdev.xml']:
        path = os.path.join(xkb_root, 'rules', filename)
        tree = etree.parse(path, etree.XMLParser(remove_blank_text=True))
        has_obsolete_mark = False

        for variant in tree.xpath('//variant[@type]'):
            variant.attrib.pop('type')
            has_obsolete_mark = True

        if has_obsolete_mark:
            try:
                tree.write(path, pretty_print=True, xml_declaration=True,
                           encoding='utf-8')
                print('... ' + path)
            except Exception as exc:
                exit_FileNotWritable(exc, path)


def list_rules(xkb_root, mask='*'):
    """ List all matching XKB layouts. """

    if mask in ('', '*'):
        locale_mask = '*'
        variant_mask = '*'
    else:
        m = mask.split('/')
        if len(m) == 2:
            locale_mask, variant_mask = m
        else:
            locale_mask = mask
            variant_mask = "*"

    kb_index = {}
    for filename in ['base.xml', 'evdev.xml']:
        tree = etree.parse(os.path.join(xkb_root, 'rules', filename))
        for variant in tree.xpath('//variant'):
            locale = variant.xpath('../../configItem/name')[0].text
            name = variant.xpath('configItem/name')[0].text
            desc = variant.xpath('configItem/description')[0].text

            if locale_mask in ('*', locale) and variant_mask in ('*', name):
                if locale not in kb_index:
                    kb_index[locale] = {}
                kb_index[locale][name] = desc

    return kb_index


###############################################################################
# Exception Handling (there must be a better way...)
#

def sys_exit(message):
    print('')
    print(message)
    sys.exit(1)


def exit_LocaleNotSupported(locale):
    sys_exit(f"Error: the `{locale}` locale is not supported.")


def exit_FileNotWritable(exception, path):
    if isinstance(exception, PermissionError):  # noqa: F821
        sys_exit('Permission denied. Are you root?')
    elif isinstance(exception, IOError):
        sys_exit(f"Error: could not write to file {path}.")
    else:
        sys_exit(f"Error: {exception}.\n{traceback.format_exc()}")



#!/usr/bin/env python3
import datetime
import os
import re
import sys
import yaml

import tomli

from .template import xkb_keymap, \
    osx_keymap, osx_actions, osx_terminators, \
    klc_keymap, klc_deadkeys, klc_dk_index, \
    web_keymap, web_deadkeys

from .utils import open_local_file, load_data, text_to_lines, lines_to_text, \
    DEAD_KEYS, ODK_ID


###
# Helpers
#


def upper_key(letter):
    if len(letter) != 1:  # dead key?
        return ' '
    custom_alpha = {
        '\u00df': '\u1e9e',  # ß ẞ
        '\u007c': '\u00a6',  # | ¦
        '\u003c': '\u2264',  # < ≤
        '\u003e': '\u2265',  # > ≥
        '\u2020': '\u2021',  # † ‡
        '\u2190': '\u21d0',  # ← ⇐
        '\u2191': '\u21d1',  # ↑ ⇑
        '\u2192': '\u21d2',  # → ⇒
        '\u2193': '\u21d3',  # ↓ ⇓
        '\u00b5': ' ',       # µ (to avoid getting `Μ` as uppercase)
    }
    if letter in custom_alpha:
        return custom_alpha[letter]
    if letter.upper() != letter.lower():
        return letter.upper()
    return ' '


def substitute_lines(text, variable, lines):
    prefix = 'KALAMINE::'
    exp = re.compile('.*' + prefix + variable + '.*')

    indent = ''
    for line in text.split('\n'):
        m = exp.match(line)
        if m:
            indent = m.group().split(prefix)[0]
            break

    return exp.sub(lines_to_text(lines, indent), text)


def substitute_token(text, token, value):
    exp = re.compile('\\$\\{' + token + '(=[^\\}]*){0,1}\\}')
    return exp.sub(value, text)


def load_tpl(layout, ext):
    tpl = 'base'
    if layout.has_altgr:
        tpl = 'full'
        if layout.has_1dk and ext.startswith('.xkb'):
            tpl = 'full_1dk'
    out = open_local_file(os.path.join('tpl', tpl + ext)).read()
    out = substitute_lines(out, 'GEOMETRY_base', layout.base)
    out = substitute_lines(out, 'GEOMETRY_full', layout.full)
    out = substitute_lines(out, 'GEOMETRY_altgr', layout.altgr)
    for key, value in layout.meta.items():
        out = substitute_token(out, key, value)
    return out


def load_descriptor(file_path):
    if file_path.endswith('.yaml') or file_path.endswith('.yml'):
        with open(file_path, encoding='utf-8') as file:
            return yaml.load(file, Loader=yaml.SafeLoader)
    with open(file_path, mode='rb') as file:
        return tomli.load(file)


###
# Constants
#


CONFIG = {
    'author': 'Fabien Cazenave',
    'license': 'WTFPL - Do What The Fuck You Want Public License',
    'geometry': 'ISO'
}

SPACEBAR = {
    'shift':       " ",  # U+0020 SPACE
    'altgr':       " ",  # U+0020 SPACE
    'altgr_shift': " ",  # U+0020 SPACE
    '1dk':         "'",  # U+0027 APOSTROPHE
    '1dk_shift':   "'"   # U+0027 APOSTROPHE
}

GEOMETRY = load_data('geometry.yaml')


###
# Main
#


class KeyboardLayout:
    """ Lafayette-style keyboard layout: base + 1dk + altgr layers. """

    def __init__(self, filepath):
        """ Import a keyboard layout to instanciate the object. """

        # initialize a blank layout
        self.layers = [{}, {}, {}, {}, {}, {}]
        self.dead_keys = {}        # dictionary subset of DEAD_KEYS
        self.dk_index = []         # ordered keys of the above dictionary
        self.meta = CONFIG.copy()  # default parameters, hardcoded
        self.has_altgr = False
        self.has_1dk = False

        # load the YAML data (and its ancessor, if any)
        try:
            cfg = load_descriptor(filepath)
            if 'extends' in cfg:
                path = os.path.join(os.path.dirname(filepath), cfg['extends'])
                ext = load_descriptor(path)
                ext.update(cfg)
                cfg = ext
        except Exception as exc:
            print('File could not be parsed.')
            print(f"Error: {exc}.")
            sys.exit(1)

        # metadata: self.meta
        for k in cfg:
            if k != 'base' and k != 'full' and k != 'altgr' \
                    and not isinstance(cfg[k], dict):
                self.meta[k] = cfg[k]
        filename = os.path.splitext(os.path.basename(filepath))[0]
        self.meta['name'] = cfg['name'] if 'name' in cfg else filename
        self.meta['name8'] = cfg['name8'] if 'name8' in cfg \
            else self.meta['name'][0:8]
        self.meta['fileName'] = self.meta['name8'].lower()
        self.meta['lastChange'] = datetime.date.today().isoformat()

        # keyboard layers: self.layers & self.dead_keys
        rows = GEOMETRY[self.meta['geometry']]['rows']
        if 'full' in cfg:
            full = text_to_lines(cfg['full'])
            self._parse_template(full, rows, 0)
            self._parse_template(full, rows, 4)
            self.has_altgr = True
        else:
            base = text_to_lines(cfg['base'])
            self._parse_template(base, rows, 0)
            self._parse_template(base, rows, 2)
            if 'altgr' in cfg:
                self.has_altgr = True
                self._parse_template(text_to_lines(cfg['altgr']), rows, 4)

        # space bar
        spc = SPACEBAR.copy()
        if 'spacebar' in cfg:
            for k in cfg['spacebar']:
                spc[k] = cfg['spacebar'][k]
        self.layers[0]['spce'] = ' '
        self.layers[1]['spce'] = spc['shift']
        self.layers[2]['spce'] = spc['1dk']
        self.layers[3]['spce'] = spc['shift_1dk'] if 'shift_1dk' in spc \
            else spc['1dk']
        if self.has_altgr:
            self.layers[4]['spce'] = spc['altgr']
            self.layers[5]['spce'] = spc['altgr_shift']

        # active dead keys: self.dk_index
        for dk in DEAD_KEYS:
            if dk['char'] in self.dead_keys:
                self.dk_index.append(dk['char'])

        # remove unused characters in self.dead_keys[].{base,alt}
        def layer_has_char(char, layer_index):
            for id in self.layers[layer_index]:
                if self.layers[layer_index][id] == char:
                    return True
            return False

        for dk_id in self.dead_keys:
            base = self.dead_keys[dk_id]['base']
            alt = self.dead_keys[dk_id]['alt']
            used_base = ''
            used_alt = ''
            for i in range(len(base)):
                if layer_has_char(base[i], 0) or layer_has_char(base[i], 1):
                    used_base += base[i]
                    used_alt += alt[i]
            self.dead_keys[dk_id]['base'] = used_base
            self.dead_keys[dk_id]['alt'] = used_alt

        # 1dk behavior
        if ODK_ID in self.dead_keys:
            self.has_1dk = True
            odk = self.dead_keys[ODK_ID]
            # alt_self (double-press), alt_space (1dk+space)
            odk['alt_space'] = spc['1dk']
            for key in self.layers[0]:
                if self.layers[0][key] == ODK_ID:
                    odk['alt_self'] = self.layers[2][key]
                    break
            # copy the 2nd and 3rd layers to the dead key
            for i in [0, 1]:
                for (name, alt_char) in self.layers[i + 2].items():
                    base_char = self.layers[i][name]
                    if name != 'spce' and base_char != ODK_ID:
                        odk['base'] += base_char
                        odk['alt'] += alt_char

    def _parse_template(self, template, rows, layer_number):
        """ Extract a keyboard layer from a template. """

        if layer_number == 0:  # base layer
            col_offset = 0
        else:  # AltGr or 1dk
            col_offset = 2

        j = 0
        for row in rows:
            i = row['offset'] + col_offset
            keys = row['keys']

            base = list(template[2 + j * 3])
            shift = list(template[1 + j * 3])

            for key in keys:
                base_key = ('*' if base[i - 1] == '*' else '') + base[i]
                shift_key = ('*' if shift[i - 1] == '*' else '') + shift[i]

                if layer_number == 0 and base_key == ' ':  # 'shift' prevails
                    base_key = shift_key.lower()
                if layer_number != 0 and shift_key == ' ':
                    shift_key = upper_key(base_key)

                if base_key != ' ':
                    self.layers[layer_number + 0][key] = base_key
                if shift_key != ' ':
                    self.layers[layer_number + 1][key] = shift_key

                for dk in DEAD_KEYS:
                    if base_key == dk['char']:
                        self.dead_keys[base_key] = dk.copy()
                    if shift_key == dk['char']:
                        self.dead_keys[shift_key] = dk.copy()

                i += 6

            j += 1

    ###
    # Geometry: base, full, altgr
    #

    def _fill_template(self, template, rows, layer_number):
        """ Fill a template with a keyboard layer. """

        if layer_number == 0:  # base layer
            col_offset = 0
            shift_prevails = True
        else:  # AltGr or 1dk
            col_offset = 2
            shift_prevails = False

        j = 0
        for row in rows:
            i = row['offset'] + col_offset
            keys = row['keys']

            base = list(template[2 + j * 3])
            shift = list(template[1 + j * 3])

            for key in keys:
                base_key = ' '
                if key in self.layers[layer_number]:
                    base_key = self.layers[layer_number][key]

                shift_key = ' '
                if key in self.layers[layer_number + 1]:
                    shift_key = self.layers[layer_number + 1][key]

                dead_base = len(base_key) == 2 and base_key[0] == '*'
                dead_shift = len(shift_key) == 2 and shift_key[0] == '*'

                if shift_prevails:
                    shift[i] = shift_key[-1]
                    if dead_shift:
                        shift[i-1] = '*'
                    if upper_key(base_key) != shift_key:
                        base[i] = base_key[-1]
                        if dead_base:
                            base[i-1] = '*'
                else:
                    base[i] = base_key[-1]
                    if dead_base:
                        base[i-1] = '*'
                    if upper_key(base_key) != shift_key:
                        shift[i] = shift_key[-1]
                        if dead_shift:
                            shift[i-1] = '*'

                i += 6

            template[2 + j * 3] = ''.join(base)
            template[1 + j * 3] = ''.join(shift)
            j += 1

        return template

    def _get_geometry(self, layers=[0], name='ISO'):
        """ `geometry` view of the requested layers. """

        rows = GEOMETRY[name]['rows']
        template = GEOMETRY[name]['template'].split('\n')[:-1]
        for i in layers:
            template = self._fill_template(template, rows, i)
        return template

    @property
    def base(self):
        """ Base + 1dk layers. """
        return self._get_geometry([0, 2])

    @property
    def full(self):
        """ Base + AltGr layers. """
        return self._get_geometry([0, 4])

    @property
    def altgr(self):
        """ AltGr layer only. """
        return self._get_geometry([4])

    ###
    # OS-specific drivers: keylayout, klc, xkb, xkb_patch
    #

    @property
    def keylayout(self):
        """ macOS driver """
        out = load_tpl(self, '.keylayout')
        for i, layer in enumerate(osx_keymap(self)):
            out = substitute_lines(out, 'LAYER_' + str(i), layer)
        out = substitute_lines(out, 'ACTIONS', osx_actions(self))
        out = substitute_lines(out, 'TERMINATORS', osx_terminators(self))
        return out

    @property
    def klc(self):
        """ Windows driver (warning: requires CR/LF + UTF16LE encoding) """
        out = load_tpl(self, '.klc')
        out = substitute_lines(out, 'LAYOUT', klc_keymap(self))
        out = substitute_lines(out, 'DEAD_KEYS', klc_deadkeys(self))
        out = substitute_lines(out, 'DEAD_KEY_INDEX', klc_dk_index(self))
        out = substitute_token(out, 'encoding', 'utf-16le')
        return out

    @property
    def xkb(self):  # will not work with Wayland
        """ GNU/Linux driver (standalone / user-space) """
        out = load_tpl(self, '.xkb')
        out = substitute_lines(out, 'LAYOUT', xkb_keymap(self, False))
        return out

    @property
    def xkb_patch(self):
        """ GNU/Linux driver (system patch) """
        out = load_tpl(self, '.xkb_patch')
        out = substitute_lines(out, 'LAYOUT', xkb_keymap(self, True))
        return out

    ###
    # JSON output: keymap (base+altgr layers) and dead keys
    #

    @property
    def json(self):
        """ JSON layout descriptor """
        return {
            'name': self.meta['name'],
            'description': self.meta['description'],
            'geometry': self.meta['geometry'].lower(),
            'keymap': web_keymap(self),
            'deadkeys': web_deadkeys(self),
            'altgr': self.has_altgr
        }



from setuptools import setup

with open('README.rst', encoding='utf-8') as f:
    readme = f.read()

setup(
    name='kalamine',
    version='0.16',
    description='a cross-platform Keyboard Layout Maker',
    long_description=readme,
    url='http://github.com/fabi1cazenave/kalamine',
    author='Fabien Cazenave',
    author_email='fabien@cazenave.cc',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
    packages=['kalamine'],
    entry_points={
        'console_scripts': [
            'kalamine = kalamine.cli:make',
            'xkalamine = kalamine.cli_xkb:cli',
        ]
    },
    install_requires=[
        'click',
        'pyyaml',
        'tomli',
        'lxml',
    ],
    include_package_data=True,
    zip_safe=False
)
